import csv
import re
from collections import OrderedDict, defaultdict, Iterable
from itertools import groupby
import pickle
from phigaro.misc.ranges import first
import numpy as np
import os

INFINITY = float('inf')


def read_npn(filename, sep=None, as_dict=True):
    sep = sep or ' '
    with open(filename) as f:
        reader = csv.reader(f, delimiter=sep)
        if as_dict:
            return dict(reader)
        else:
            return reader


def read_coords(filename, sep=None):
    sep = sep or '\t'
    with open(filename) as f:
        reader = csv.reader(f, delimiter=sep)
        res = {}
        for phage, group in groupby(reader, key=lambda x: x[0]):
            res[phage] = sorted((
                (int(begin), int(end))
                for _, begin, end in group
            ), key=lambda x: x[0])
    return res


def convert_npn(phage, ph_sym):
    return [
        int(c == ph_sym)
        for c in phage
    ]


def hmmer_res_to_npn(scaffold, hmmer_result, max_evalue, penalty_black, penalty_white, pvogs_black_list, pvogs_white_list):
    # type: (Scaffold, HmmerResult, float)->list[int]
    ordered_records_it = (
        HmmerResult.min_record(hmmer_result.get_records(scaffold.name, gene.name))
        for gene in scaffold
    )

    return [
        0
        if not record or record.evalue > max_evalue
        else 1 - penalty_black if record.vog_name in pvogs_black_list
        else 1 + penalty_white if record.vog_name in pvogs_white_list
        else 1
        for record in ordered_records_it
    ]

def hmmer_res_to_gc(scaffold, hmmer_result, max_evalue):
    # type: (Scaffold, HmmerResult, float)->list[int]
    ordered_records_it = (
        HmmerResult.min_record(hmmer_result.get_records(scaffold.name, gene.name))
        for gene in scaffold
    )

    return [
        0
        if not record or record.evalue > max_evalue
        else record.gc_cont
        for record in ordered_records_it
    ]


class Gene(object):
    def __init__(self, name, begin, end, scaffold=None):
        # type: (str, int, int, Scaffold|None)->Gene
        self.name = name
        self.begin = begin
        self.end = end
        self.scaffold = scaffold


class Scaffold(object):
    def __init__(self, name, genes):
        # type: (str, list[Gene])->Scaffold
        self.name = name
        self._genes_map = OrderedDict()

        for gene in sorted(genes, key=lambda g: g.begin):
            gene.scaffold = self
            self._genes_map[gene.name] = gene

    def __iter__(self):
        # type: ()->Iterable[Gene]
        return iter(self._genes_map.values())

    def get_gene(self, gene_name):
        # type: (str)->Gene
        return self._genes_map[gene_name]


class ScaffoldSet(object):
    def __init__(self, scaffolds):
        # type: (list[Scaffold])->ScaffoldSet
        self._scaffolds_map = {
            scaffold.name: scaffold
            for scaffold in scaffolds
        }

    def get_scaffold(self, scaffold_name):
        # type: (str)->Scaffold
        return self._scaffolds_map[scaffold_name]

    def __iter__(self):
        return iter(self._scaffolds_map.values())


class HmmerRecord(object):
    def __init__(self, scaffold_name, gene_name, vog_name, evalue, gc_cont, begin, end, strand):
        # type: (str, str, str, float)->HmmerRecord
        self.scaffold_name = scaffold_name
        self.gene_name = gene_name
        self.vog_name = vog_name
        self.evalue = evalue
        self.gc_cont = gc_cont
        self.begin = begin
        self.end = end
        self.strand = strand

class HmmerResult(object):
    def __init__(self, hmmer_records):
        # type: (Iterable[HmmerRecord])->HmmerResult
        self._scaffolds_map = defaultdict(lambda: defaultdict(list))  # type: dict[str, dict[str, list[HmmerRecord]]]
        for hmmer_record in hmmer_records:
            self._add_record(hmmer_record)

    def _add_record(self, hmmer_record):
        # type: (HmmerRecord)->None
        scaffold_records = self._scaffolds_map[hmmer_record.scaffold_name]
        scaffold_records[hmmer_record.gene_name].append(hmmer_record)

    def get_records(self, scaffold_name, gene_name):
        # type: (str, str)->list[HmmerRecord]
        return self._scaffolds_map[scaffold_name].get(gene_name, [])

    @staticmethod
    def min_record(records):
        # type: (list[HmmerRecord])->HmmerRecord|None
        if not records:
            return None
        return min(records, key=lambda r: r.evalue)


def read_hmmer_output(file_path):
    # type: (str)->HmmerResult

    def parse_line(line):
        # type: (str)->HmmerRecord
        tokens = re.split(r'\s+', line)
        begin, end, strand = line.split(' # ', 4)[1:-1]
        gc_cont = float([x.split('=')[-1] for x in tokens[-1].split(';') if x.startswith('gc_cont')][0])

        return HmmerRecord(
            scaffold_name='_'.join(tokens[0].split('_')[:-1]),
            gene_name=tokens[0].split('_')[-1],
            vog_name=tokens[2],
            evalue=float(tokens[4]),
            gc_cont=gc_cont,
            begin = int(begin),
            end = int(end),
            strand = int(strand)
        )

    with open(file_path) as f:
        lines_it = (
            parse_line(line.strip())
            for line in f
            if not line.startswith('#') and line.strip()
        )
        hmm_res = HmmerResult(lines_it)

        return hmm_res


def read_prodigal_output(file_name):
    # type: (str)->ScaffoldSet

    def extract_coords_and_name(gene_str):
        # type: (str)->Gene

        tokens = gene_str.split(' # ')
        return Gene(
            name=tokens[0],
            begin=int(tokens[1]),
            end=int(tokens[2])
        )

    def parse_gene_records(gene_records):
        # type: (list[str])->list[Gene]
        return [
            extract_coords_and_name(gene_str)
            for _, gene_str in gene_records
        ]
    with open(file_name) as f:
        genes_scaffold_recs = (
            ['_'.join(line.strip().split(' # ', 1)[0].split('_')[:-1]),
             line.strip().split(' # ', 1)[0].split('_')[-1]+' # '+line.strip().split(' # ', 1)[1]]
            for line in f
            if line.startswith('>')
        )
        scaffolds = [
            Scaffold(
                name=scaffold_name[1:],
                genes=parse_gene_records(gene_records)
            )
            for scaffold_name, gene_records in groupby(genes_scaffold_recs, key=first)
        ]
        return ScaffoldSet(scaffolds=scaffolds)

def define_taxonomy(pvogs_string):
    with open(os.path.dirname(os.path.abspath(__file__))+'/pvogs_taxonomy.pickle', 'rb') as f:
        pvogs_taxonomy, taxonomy_codes = pickle.load(f)
    taxonomy_codes = taxonomy_codes.astype(str)
    pvog_list = pvogs_string.split(', ')
    taxonomy_variants = [pvogs_taxonomy[pvog] for pvog in pvog_list if pvog in pvogs_taxonomy]
    if len(taxonomy_variants) > 0:
        taxonomies, counts = np.unique(taxonomy_variants, return_counts=True)
        return ' / '.join(taxonomy_codes[taxonomies[counts == max(counts)]])
    else:
        return 'Unknown'




import csv
from builtins import super

from phigaro.data import read_prodigal_output, read_hmmer_output, hmmer_res_to_npn, hmmer_res_to_gc, Gene, define_taxonomy
from phigaro.finder.v2 import V2Finder
from phigaro import const
from .base import AbstractTask
from .prodigal import ProdigalTask
from .hmmer import HmmerTask
import os

from phigaro.to_html.preprocess import plot_html, form_sequence, if_transposable
from phigaro.to_html.html_formation import form_html_document

class RunPhigaroTask(AbstractTask):
    task_name = 'run_phigaro'

    def __init__(self, hmmer_task, prodigal_task):
        """
        :type hmmer_task: HmmerTask
        :type prodigal_task: ProdigalTask
        """
        super().__init__()
        self.hmmer_task = hmmer_task
        self.prodigal_task = prodigal_task

    def _prepare(self):
        self.finder = V2Finder(
            window_len=self.config['phigaro']['window_len'],
            threshold_min=self.config['phigaro']['threshold_min'],
            threshold_max=self.config['phigaro']['threshold_max'],
        )
        self._print_vogs = self.config['phigaro'].get('print_vogs', False)
        self._filename = self.config['phigaro'].get('filename', False)
        self._no_html = self.config['phigaro'].get('no_html', False)
        self._not_open = self.config['phigaro'].get('not_open', False)
        self._output = self.config['phigaro'].get('output', False)
        self._uuid = self.config['phigaro'].get('uuid', False)

    def output(self):
        return self.file('{}.tsv'.format(self.sample))

    def run(self):
        max_evalue = self.config['hmmer']['e_value_threshold']
        penalty_black = self.config['phigaro']['penalty_black']
        penalty_white = self.config['phigaro']['penalty_white']

        pvogs_black_list = const.DEFAULT_BLACK_LIST
        pvogs_white_list = const.DEFAULT_WHITE_LIST

        scaffold_set = read_prodigal_output(self.prodigal_task.output())
        hmmer_result = read_hmmer_output(self.hmmer_task.output())

        with open(self.output(), 'w') as of:
            writer = csv.writer(of, delimiter='\t')

            if self._print_vogs:
                writer.writerow(('scaffold', 'begin', 'end', 'transposable', 'taxonomy', 'vog'))
            else:
                writer.writerow(('scaffold', 'begin', 'end', 'transposable', 'taxonomy'))

            plotly_plots = []
            phage_info = []
            transposables_status = []
            phage_num = 0
            for scaffold in scaffold_set:
                phage_info.append([scaffold.name, []])
                genes = list(scaffold)  # type: list[Gene]
                npn = hmmer_res_to_npn(scaffold, hmmer_result, max_evalue=max_evalue,
                                       penalty_black = penalty_black, penalty_white = penalty_white,
                                       pvogs_black_list = pvogs_black_list, pvogs_white_list = pvogs_white_list)
                gc = hmmer_res_to_gc(scaffold, hmmer_result, max_evalue=max_evalue)

                phages = self.finder.find_phages(npn, gc)
                for phage in phages:
                    phage_num += 1
                    begin = genes[phage.begin].begin
                    end = genes[phage.end].end
                    hmmer_records = [
                        hmmer_result.min_record(hmmer_result.get_records(scaffold.name, gene.name))
                        for gene in genes[phage.begin: phage.end]
                    ]
                    hmmer_pvogs_records = (
                        record.vog_name
                        for record in hmmer_records
                        if record and record.evalue <= max_evalue
                    )
                    pvogs_records_str = ', '.join(hmmer_pvogs_records)
                    taxonomy = define_taxonomy(pvogs_records_str)
                    hmmer_records = [
                        record
                        for record in hmmer_records
                        if record and record.evalue <= max_evalue
                    ]
                    transposable = if_transposable(hmmer_records)
                    if self._print_vogs:
                        writer.writerow((scaffold.name, begin, end, transposable, taxonomy,  pvogs_records_str))
                    else:
                        writer.writerow((scaffold.name, begin, end, transposable, taxonomy))

                    sequence = form_sequence(self._filename, '%s_prophage_%d' % (scaffold.name, phage_num), begin, end, scaffold.name)
                    transposables_status.append(transposable)
                    the_phage_info = [begin, end, taxonomy, sequence, pvogs_records_str]
                    phage_info[-1][1].append(the_phage_info)
                    if (not self._no_html) and (self._output):
                        plotly_plots.append(plot_html(hmmer_records, begin, end))
                phage_info = phage_info if len(phage_info[-1][1]) > 0 else phage_info[:-1]
            if (len(phage_info) > 0) and (not self._no_html) and (self._output):
                html = form_html_document(phage_info, transposables_status, plotly_plots, self._filename, self._uuid)
                with open(self._output+'.html', 'w') as f:
                    f.write(html)
                if not self._not_open:
                    os.system('xdg-open '+self._output+'.html'+' > /dev/null 2>/dev/null')


from builtins import super

from .base import AbstractTask
import sh


class ProdigalTask(AbstractTask):
    task_name = 'prodigal'

    def __init__(self, preprocess_task):
        super().__init__()

        self.preprocess_task = preprocess_task
        self.prodigal = (
            sh.Command(self.config['prodigal']['bin'])
            .bake(
                i=self.preprocess_task.output(),
                a=self.output()
            )
        )

    def output(self):
        return self.file('{}.faa'.format(self.sample))

    def run(self):
        self.prodigal()


from .html_templates import *
from bs4 import BeautifulSoup

def form_html_document(prophage_data, transposables_status, plots_data, filename_input, uuid):
  
    def form_tbody(data, transposables_data):
        prophage_index = 1
        tbody = BeautifulSoup(features="lxml").new_tag('tbody')
        for scaffold, scaffold_info in data:
            tr = BeautifulSoup(features="lxml").new_tag('tr', **{'class':'scaffold-name'})
            th = BeautifulSoup(features="lxml").new_tag('th', scope="row", colspan="4")
            th.append(scaffold)
            tr.append(th)
            tbody.append(tr)
            for prophage_info in scaffold_info:
                index_to_paste = BeautifulSoup(transposable_index.format(prophage_index), 'html.parser') if transposables_data[prophage_index-1] else prophage_index
                if (prophage_index == 1):
                    prophage_info = [prophage_index]+prophage_info+['active', index_to_paste]
                else:
                    prophage_info = [prophage_index]+prophage_info+['', index_to_paste]
                tbody.append(BeautifulSoup(row.format(*prophage_info), 'html.parser'))
                prophage_index += 1
        return str(tbody)
                
    def form_plots_body(data):
        plots_body = BeautifulSoup(features="lxml")
        for prophage_index, plot in enumerate(data):
            if prophage_index == 0 :
                div = BeautifulSoup(features="lxml").new_tag('div', **{"class": "tab-pane fade show active",
                                                    "id":"phage-%d"%(prophage_index+1)})
                inner_div = BeautifulSoup(features="lxml").new_tag('div', **{"class": "prophage-plot"})
            else:
                div = BeautifulSoup(features="lxml").new_tag('div', **{"class": "tab-pane fade show",
                                                                       "id": "phage-%d" % (prophage_index + 1)})
                inner_div = BeautifulSoup(features="lxml").new_tag('div', **{"class": "prophage-plot", "style":"display:none"})
            inner_div.append(BeautifulSoup(plot, 'html.parser'))
            div.append(inner_div)
            plots_body.append(div)
        return str(plots_body)
    
    def arrange_html_parts(prophage_data, transposables_status, plots_data, filename_input, uuid):
        global header, style_css, body_main, body_table, body_plots, row, footer
        header = header.format(filename_input)
        tbody = form_tbody(prophage_data, transposables_status)
        plots_body = form_plots_body(plots_data)
        body_table = body_table.format(tbody)
        body_plots = body_plots.format(plots_body)
        body_main = body_main.format(body_table, body_plots)
        footer = footer.replace('<script type="text/javascript">','<script type="text/javascript">\n    var uuid="%s";'%uuid)
        html = header + style_css + body_main + footer
        return html
    
    return arrange_html_parts(prophage_data, transposables_status, plots_data, filename_input, uuid)


from singleton.singleton import Singleton


@Singleton
class Context(object):
    def __init__(self, sample, threads, config):
        """

        :type sample: str
        :type config: dict
        :type threads: int
        """
        self.threads = threads
        self.sample = sample
        self.config = config



import numpy as np
from plotly import tools
from plotly.graph_objs import Bar
import plotly.offline as py

from phigaro.finder.v2 import calc_scores

def _make_coords_colors(data_len, real_phage_coords):
    colors = np.zeros(data_len)
    for begin, end in real_phage_coords:
        for i in range(begin, end + 1):
            colors[i] = 1
    return colors


def plot_scores(scores, title, real_phage_coords=None):
    indices = np.arange(len(scores))

    colors = None
    if real_phage_coords is not None:
        colors = _make_coords_colors(len(scores), real_phage_coords)
    data = Bar(
        x=indices,
        y=scores + 0.1,
        name=title,
        marker=dict(
            color=colors,
        )
    )

    return data


def plot_phage(phage, title):
    ind = np.arange(len(phage))
    int_phage = [c + .1 for c in phage]
    data = Bar(
        x=ind,
        y=int_phage,
        marker=dict(
            color='black',
        ),
        name=title
    )
    return data


def _make_rects(coords, ymin, ymax, fillcolor, opacity):
    return [
        dict(
            type='rect',
            xref='x',
            yref='y',
            x0=x_begin,
            y0=ymin,
            x1=x_end,
            y1=ymax,
            fillcolor=fillcolor,
            opacity=opacity,
            line={'width': 0}
        )
        for (x_begin, x_end) in coords
        ]


def plot_scores_and_phage(phage, window_len, score_func=None, scan_func=None, real_phage_coords=None):
    score_func = score_func or score_tri
    fig = tools.make_subplots(rows=2, cols=1, shared_xaxes=True)
    title = 'Scores: window: {}'.format(window_len)
    scores = np.array(calc_scores(phage, window_len, score_func))

    ranges = []
    if scan_func is not None:
        ranges = scan_func(scores)

    score_fig = plot_scores(scores, title, real_phage_coords=None)
    phage_fig = plot_phage(phage, 'Phage')

    fig.append_trace(score_fig, 1, 1)
    fig.append_trace(phage_fig, 2, 1)

    ymax = window_len / 2
    if real_phage_coords is not None or ranges:
        fig['layout'].update(dict(
            shapes=_make_rects(ranges, ymax, 'rgb(50, 171, 96)', 0.5) + _make_rects(real_phage_coords or [], ymax, '#ff0000', 0.5)

        ))

    py.iplot(fig)


def plot_scores_and_phage2(phage, scores, found_phage_coords, real_phage_coords=None, filename='filename'):
    # real_phage_coords = real_phage_coords or []
    fig = tools.make_subplots(rows=2, cols=1, shared_xaxes=True)
    title = 'Scores'

    score_fig = plot_scores(scores, title, real_phage_coords=None)
    phage_fig = plot_phage(phage, 'Phage')

    fig.append_trace(score_fig, 1, 1)
    fig.append_trace(phage_fig, 2, 1)

    ymax = max(scores)
    # print(len(real_phage_coords), len(found_phage_coords))
    if (len(real_phage_coords) + len(found_phage_coords)) != 0:
        # print('got real coords')
        fig['layout'].update(dict(
            shapes=_make_rects(found_phage_coords, ymax * 0.5, ymax * 0.75, '#0000ff', 0.5) + \
                   _make_rects(real_phage_coords, ymax * 0.75, ymax, '#aaaa00', 0.5)

        ))

    py.plot(fig, filename=filename+'.html')



import pickle
import plotly.graph_objs as go
from plotly.offline import plot
import numpy as np
import os
from Bio import SeqIO

pvogs_groups = ['Lysis', 'Integration', 'Terminase', 'Replication',
                                'Coat', 'Baseplate', 'Tail', 'Assembly','Portal','Other (structural)','Other']
cmap = [(31,119,180), (255,127,14), (44,160,44), (214,39,40),
        (188,189,34), (23,190,207), (148,103,189), (227,119,194),
        (0,0,128),(32,50,50),(127,127,127)]
colors = {group:cmap[i] for i, group in enumerate(pvogs_groups)}

with open(os.path.dirname(os.path.abspath(__file__))+'/pvogs_annotations.pickle', 'rb') as f:
    annotations = pickle.load(f)

def form_arrow(begin, end, nonreverse, width):
    '''
    :param begin: int (of gene)
    :param end: int (of gene)
    :param nonreverse: boolean
    :param width:
    :return: tuple of tuples (arrow coords)
    '''
    k = 0.05
    height = 0.9 if width < 0.4 else 1
    width = width if width > 0.6 else 0.6
    if nonreverse:
        return ((begin, 1-width*k), (begin, 1+width*k), (begin+0.6*(end-begin), 1+width*k),
                (begin+0.6*(end-begin), 1+1.8*height*k), (end, 1),
                (begin+0.6*(end-begin), 1-1.8*height*k), (begin+0.6*(end-begin), 1-width*k))
    else:
        return ((end, 1-width*k), (end, 1+width*k), (end-0.4*(end-begin), 1+width*k),
                (end-0.4*(end-begin), 1+1.8*height*k), (begin, 1),
                (end-0.4*(end-begin), 1-1.8*height*k), (end-0.4*(end-begin), 1-width*k))

def plot_html(records, prophage_begin, prophage_end):
    '''
    :param records: list of Gene
    :return: html_plot
    '''
    def _plotly_to_html_(prophage_begin, prophage_end, func_groups, widths, coords, arrow_coords):
        global colors, pvogs_groups
        traces0 = [go.Scatter(
            x=[prophage_begin],
            y=[100],
            mode='markers',
            opacity=1,
            hoverinfo='none',
            marker=dict(
                size=10,
                color='rgba(%.5f, %.5f, %.5f, 0.8)' % colors[group],
            ),
            name=group
        )
        for group in pvogs_groups if group in func_groups.keys()
        ] + [go.Scatter(
            x=np.array(coords)[elements],
            y=[1] * len(elements),
            text=np.array(records_info)[elements],
            hoverinfo='text',
            mode='markers',
            marker=dict(
                size=90 * np.array(widths)[elements],
                color='rgba(%.5f, %.5f, %.5f, 1)' % colors[group],
                opacity=0
            ),
            showlegend=False
        )
        for group, elements in func_groups.items()]
        trace1 = [go.Scatter(
            x=[prophage_begin, prophage_end],
            y=[1, 1],
            mode='lines',
            line=dict(
                color='rgba(116,116,116, 0.5)',
                width=1
            ),
            hoverinfo='none',
            showlegend=False
        )]
        layout = {
            'yaxis': dict(
                range=[-.5, 2.5],
                showgrid=False,
                zeroline=False,
                showline=False,
                ticks='',
                showticklabels=False
            ),
            'xaxis': dict(
                title='Nucleotide Number'
            ),
            'shapes': [
                {
                    'type': 'path',
                    'path': 'M ' + ' '.join(['L {:.5f},{:.5f}'.format(*arrow_point) for arrow_point in arrow])[1:] + ' Z',
                    'fillcolor': 'rgba(%.5f, %.5f, %.5f, 1)' % colors[group],  # 'rgba(44, 160, 101, 0.5)'
                    'opacity': 0.7,
                    'line': {
                        'color': 'rgba(%.5f, %.5f, %.5f, 0.7)' % colors[group],
                    },
                }
                for group, elements in func_groups.items() for arrow in np.array(arrow_coords)[elements]]
        }
        data = traces0 + trace1
        fig = {
            'data': data,
            'layout': layout,
        }
        return fig

    pvogs = []
    widths = []
    coords = []
    arrow_coords = []
    records_info = []
    func_groups = {}
    for i, record in enumerate(records):
        pvogs.append(record.vog_name)
        widths.append(record.end - record.begin)
        coords.append(record.begin + (record.end - record.begin) / 2.0)
        arrow_coords.append((record.begin, record.end, (record.strand + 1) / 2))
        records_info.append('%s<br>%d-%d<br>gc_cont = %s' % (record.vog_name, record.begin, record.end, record.gc_cont))
        group = annotations.loc[record.vog_name.strip()].group if record.vog_name.strip() in annotations.index else 'Other'
        if group in func_groups.keys():
            func_groups[group].append(i)
        else:
            func_groups[group] = [i]
    widths = 1.0 * (np.array(widths) - min(widths)) / max(widths) + 0.1
    arrow_coords = [form_arrow(*info, width=width) for info, width in zip(arrow_coords, widths)]
    plotly_fig = _plotly_to_html_(prophage_begin, prophage_end, func_groups, widths, coords, arrow_coords)
    return plot(plotly_fig, include_plotlyjs=False, filename='gene_map.html', output_type='div')

def form_sequence(fasta_file, fasta_name, prophage_begin, prophage_end, scaffold_name):
    for record in SeqIO.parse(fasta_file, "fasta"):
        if record.id.strip() == scaffold_name.strip():
            sequence = record.seq[prophage_begin:(prophage_end + 1)]
            ins_ind = 77
            while ins_ind < len(sequence):
                sequence = sequence[:ins_ind] + '%0A' + sequence[ins_ind:]
                ins_ind += 80
            sequence = '%' + '3E%s' % fasta_name + '%' + '0A%s' % sequence
            return sequence

def if_transposable(records):
    status = False
    for record in records:
        group = annotations.loc[record.vog_name.strip()].group if record.vog_name.strip() in annotations.index else 'Other'
        if (group == 'Integration') & status:
            return True
        elif (group == 'Integration'):
            status = True
        elif (group != 'Integration') & (group != 'Other'):
            status = False
    return False
